import sys
import os
import glob
import numpy
import time

from PyMca5.PyMca import specfilewrapper, EdfFile, SixCircle, specfile

from .. import backend, errors, util


class pixels(backend.ProjectionBase):
    def project(self, wavelength, UB, gamma, delta, theta, mu, chi, phi):
        y, x = numpy.mgrid[slice(None, gamma.shape[0]), slice(None, delta.shape[0])]
        return (y, x)

    def get_axis_labels(self):
        return "y", "x"


class HKLProjection(backend.ProjectionBase):
    # arrays: gamma, delta
    # scalars: theta, mu, chi, phi
    def project(self, wavelength, UB, beta, delta, omega, alfa, chi, phi):
        R = SixCircle.getHKL(
            wavelength,
            UB,
            gamma=beta,
            delta=delta,
            theta=omega,
            mu=alfa,
            chi=chi,
            phi=phi,
        )
        shape = beta.size, delta.size
        H = R[0, :].reshape(shape)
        K = R[1, :].reshape(shape)
        L = R[2, :].reshape(shape)
        return (H, K, L)

    def get_axis_labels(self):
        return "H", "K", "L"


class HKProjection(HKLProjection):
    def project(self, wavelength, UB, beta, delta, omega, alfa, chi, phi):
        H, K, L = super(HKProjection, self).project(
            wavelength, UB, beta, delta, omega, alfa, chi, phi
        )
        return (H, K)

    def get_axis_labels(self):
        return "H", "K"


class ThetaLProjection(backend.ProjectionBase):
    # arrays: gamma, delta
    # scalars: theta, mu, chi, phi
    def project(self, wavelength, UB, gamma, delta, theta, mu, chi, phi):
        R = SixCircle.getHKL(
            wavelength,
            UB,
            gamma=gamma,
            delta=delta,
            theta=theta,
            mu=mu,
            chi=chi,
            phi=phi,
        )
        shape = gamma.size, delta.size
        L = R[2, :].reshape(shape)
        theta_array = numpy.ones_like(L) * theta
        return (theta_array, L)

    def get_axis_labels(self):
        return "Theta", "L"


class QProjection(backend.ProjectionBase):
    def project(self, wavelength, UB, gamma, delta, theta, mu, chi, phi):
        shape = gamma.size, delta.size
        sixc = SixCircle.SixCircle()
        sixc.setLambda(wavelength)
        sixc.setUB(UB)
        R = sixc.getQSurface(
            gamma=gamma, delta=delta, theta=theta, mu=mu, chi=chi, phi=phi
        )
        qz = R[0, :].reshape(shape)
        qy = R[1, :].reshape(shape)
        qx = R[2, :].reshape(shape)
        return (qz, qy, qx)

    def get_axis_labels(self):
        return "qx", "qy", "qz"


class SphericalQProjection(QProjection):
    def project(self, wavelength, UB, gamma, delta, theta, mu, chi, phi):
        qz, qy, qx = super(SphericalQProjection, self).project(
            wavelength, UB, gamma, delta, theta, mu, chi, phi
        )
        q = numpy.sqrt(qx ** 2 + qy ** 2 + qz ** 2)
        theta = numpy.arccos(qz / q)
        phi = numpy.arctan2(qy, qx)
        return (q, theta, phi)

    def get_axis_labels(self):
        return "Q", "Theta", "Phi"


class CylindricalQProjection(QProjection):
    def project(self, wavelength, UB, gamma, delta, theta, mu, chi, phi):
        qz, qy, qx = super(CylindricalQProjection, self).project(
            wavelength, UB, gamma, delta, theta, mu, chi, phi
        )
        qpar = numpy.sqrt(qx ** 2 + qy ** 2)
        phi = numpy.arctan2(qy, qx)
        return (qpar, qz, phi)

    def get_axis_labels(self):
        return "qpar", "qz", "Phi"


class nrQProjection(backend.ProjectionBase):
    def project(self, wavelength, UB, gamma, delta, theta, mu, chi, phi):
        k0 = 2 * numpy.pi / wavelength
        delta, gamma = numpy.meshgrid(delta, gamma)
        mu *= numpy.pi / 180
        delta *= numpy.pi / 180
        gamma *= numpy.pi / 180

        qy = k0 * (
            numpy.cos(gamma) * numpy.cos(delta) - numpy.cos(mu)
        )  ## definition of qx, and qy same as spec at theta = 0
        qx = k0 * (numpy.cos(gamma) * numpy.sin(delta))
        qz = k0 * (numpy.sin(gamma) + numpy.sin(mu))
        return (qx, qy, qz)

    def get_axis_labels(self):
        return "qx", "qy", "qz"


class TwoThetaProjection(SphericalQProjection):
    def project(self, wavelength, UB, gamma, delta, theta, mu, chi, phi):
        q, theta, phi = super(TwoThetaProjection, self).project(
            wavelength, UB, gamma, delta, theta, mu, chi, phi
        )
        return (
            2 * numpy.arcsin(q * wavelength / (4 * numpy.pi)) / numpy.pi * 180,
        )  # note: we need to return a 1-tuple?

    def get_axis_labels(self):
        return "TwoTheta"


class Qpp(nrQProjection):
    def project(self, wavelength, UB, gamma, delta, theta, mu, chi, phi):
        qx, qy, qz = super(Qpp, self).project(
            wavelength, UB, gamma, delta, theta, mu, chi, phi
        )
        qpar = numpy.sqrt(qx ** 2 + qy ** 2)
        qpar[numpy.sign(qx) == -1] *= -1
        return (qpar, qz)

    def get_axis_labels(self):
        return "Qpar", "Qz"


class GammaDeltaTheta(
    HKLProjection
):  # just passing on the coordinates, makes it easy to accurately test the theta correction
    def project(self, wavelength, UB, gamma, delta, theta, mu, chi, phi):
        delta, gamma = numpy.meshgrid(delta, gamma)
        theta = theta * numpy.ones_like(delta)
        return (gamma, delta, theta)

    def get_axis_labels(self):
        return "Gamma", "Delta", "Theta"


class GammaDelta(
    HKLProjection
):  # just passing on the coordinates, makes it easy to accurately test the theta correction
    def project(self, wavelength, UB, gamma, delta, theta, mu, chi, phi):
        delta, gamma = numpy.meshgrid(delta, gamma)
        return (gamma, delta)

    def get_axis_labels(self):
        return "Gamma", "Delta"


class GammaDeltaMu(
    HKLProjection
):  # just passing on the coordinates, makes it easy to accurately test the theta correction
    def project(self, wavelength, UB, gamma, delta, theta, mu, chi, phi):
        delta, gamma = numpy.meshgrid(delta, gamma)
        mu = mu * numpy.ones_like(delta)
        return (gamma, delta, mu)

    def get_axis_labels(self):
        return "Gamma", "Delta", "Mu"


class BM32Input(backend.InputBase):
    # OFFICIAL API

    dbg_scanno = None
    dbg_pointno = None

    def generate_jobs(self, command):
        scans = util.parse_multi_range(",".join(command).replace(" ", ","))
        if not len(scans):
            sys.stderr.write("error: no scans selected, nothing to do\n")
        for scanno in scans:
            util.status("processing scan {0}...".format(scanno))
            scan = self.get_scan(scanno)
            if self.config.pr:
                pointcount = self.config.pr[1] - self.config.pr[0] + 1
                start = self.config.pr[0]
            else:
                start = 0
                try:
                    pointcount = scan.lines()
                except specfile.error:  # no points
                    continue
            next(self.get_images(scan, 0, pointcount - 1, dry_run=True))  # dryrun
            if pointcount > self.config.target_weight * 1.4:
                for s in util.chunk_slicer(pointcount, self.config.target_weight):
                    yield backend.Job(
                        scan=scanno,
                        firstpoint=start + s.start,
                        lastpoint=start + s.stop - 1,
                        weight=s.stop - s.start,
                    )
            else:
                yield backend.Job(
                    scan=scanno,
                    firstpoint=start,
                    lastpoint=start + pointcount - 1,
                    weight=pointcount,
                )

    def process_job(self, job):
        super(BM32Input, self).process_job(job)
        scan = self.get_scan(job.scan)
        self.metadict = dict()
        try:
            scanparams = self.get_scan_params(scan)  # wavelength, UB
            pointparams = self.get_point_params(
                scan, job.firstpoint, job.lastpoint
            )  # 2D array of diffractometer angles + mon + transm
            images = self.get_images(scan, job.firstpoint, job.lastpoint)  # iterator!

            for pp, image in zip(pointparams, images):
                yield self.process_image(scanparams, pp, image)
            util.statuseol()
        except Exception as exc:
            # exc.args = errors.addmessage(exc.args, ', An error occured for scan {0} at point {1}. See above for more information'.format(self.dbg_scanno, self.dbg_pointno))
            raise
        self.metadata.add_section("id03_backend", self.metadict)

    def parse_config(self, config):
        super(BM32Input, self).parse_config(config)
        self.config.xmask = util.parse_multi_range(
            config.pop("xmask", None)
        )  # Optional, select a subset of the image range in the x direction. all by default
        self.config.ymask = util.parse_multi_range(
            config.pop("ymask", None)
        )  # Optional, select a subset of the image range in the y direction. all by default
        self.config.specfile = config.pop("specfile")  # Location of the specfile
        self.config.imagefolder = config.pop(
            "imagefolder", None
        )  # Optional, takes specfile folder tag by default
        self.config.pr = util.parse_tuple(
            config.pop("pr", None), length=2, type=int
        )  # Optional, all range by default
        self.config.background = config.pop(
            "background", None
        )  # Optional, if supplied a space of this image is constructed
        if self.config.xmask is None:
            self.config.xmask = slice(None)
        if self.config.ymask is None:
            self.config.ymask = slice(None)
        self.config.maskmatrix = load_matrix(
            config.pop("maskmatrix", None)
        )  # Optional, if supplied pixels where the mask is 0 will be removed
        self.config.sdd = config.pop("sdd", None)  # sample to detector distance (mm)
        if self.config.sdd is not None:
            self.config.sdd = float(self.config.sdd)
        self.config.pixelsize = util.parse_tuple(
            config.pop("pixelsize", None), length=2, type=float
        )  # pixel size x/y (mm) (same dimension as sdd)

    def get_destination_options(self, command):
        if not command:
            return False
        command = ",".join(command).replace(" ", ",")
        scans = util.parse_multi_range(command)
        return dict(
            first=min(scans),
            last=max(scans),
            range=",".join(str(scan) for scan in scans),
        )

    # CONVENIENCE FUNCTIONS
    _spec = None

    def get_scan(self, scannumber):
        if self._spec is None:
            self._spec = specfilewrapper.Specfile(self.config.specfile)
        return self._spec.select("{0}.1".format(scannumber))

    def find_edfs(self, pattern):
        files = glob.glob(pattern)
        ret = {}
        for file in files:
            try:
                filename = os.path.basename(file).split(".")[0]
                imno = int(filename.split("_")[-1].split("-")[-1])
                ret[imno] = file
            except ValueError:
                continue
        return ret

    @staticmethod
    def apply_mask(data, xmask, ymask):
        roi = data[ymask, :]
        return roi[:, xmask]

    # MAIN LOGIC
    def get_scan_params(self, scan):
        self.dbg_scanno = scan.number()
        UB = numpy.array(scan.header("G")[2].split(" ")[-9:], dtype=numpy.float)
        wavelength = float(scan.header("G")[1].split(" ")[-1])

        self.metadict["UB"] = UB
        self.metadict["wavelength"] = wavelength

        return wavelength, UB

    def get_images(self, scan, first, last, dry_run=False):
        imagenos = (
            numpy.array(scan.datacol("img")[slice(first, last + 1)], dtype=numpy.int)
            + 1
        )  ##error in spec?!
        if self.config.background:
            if not os.path.exists(self.config.background):
                raise errors.FileError(
                    "could not find background file {0}".format(self.config.background)
                )
            if dry_run:
                yield
            else:
                edf = EdfFile.EdfFile(self.config.background)
                for i in range(first, last + 1):
                    self.dbg_pointno = i
                    yield edf
        else:
            try:
                uccdtagline = scan.header("M")[0].split()[-1]
                UCCD = os.path.dirname(uccdtagline).split(os.sep)
            except:
                print(
                    "warning: UCCD tag not found, use imagefolder for proper file specification"
                )
                UCCD = []
            pattern = self._get_pattern(UCCD)
            matches = self.find_edfs(pattern)
            if not set(imagenos).issubset(set(matches.keys())):
                raise errors.FileError(
                    "incorrect number of matches for scan {0} using pattern {1}".format(
                        scan.number(), pattern
                    )
                )
            if dry_run:
                yield
            else:
                for i in imagenos:
                    self.dbg_pointno = i
                    edf = EdfFile.EdfFile(matches[i])
                    yield edf

    def _get_pattern(self, UCCD):
        imagefolder = self.config.imagefolder
        if imagefolder:
            try:
                imagefolder = imagefolder.format(UCCD=UCCD, rUCCD=list(reversed(UCCD)))
            except Exception as e:
                raise errors.ConfigError(
                    "invalid 'imagefolder' specification '{0}': {1}".format(
                        self.config.imagefolder, e
                    )
                )
            else:
                if not os.path.exists(imagefolder):
                    raise errors.ConfigError(
                        "invalid 'imagefolder' specification '{0}'. Path {1} does not exist".format(
                            self.config.imagefolder, imagefolder
                        )
                    )
        else:
            imagefolder = os.path.join(*UCCD)
            if not os.path.exists(imagefolder):
                raise errors.ConfigError(
                    "invalid UCCD tag '{0}'. The UCCD tag in the specfile does not point to an existing folder. Specify the imagefolder in the configuration file.".format(
                        imagefolder
                    )
                )
        return os.path.join(imagefolder, "*")


class EH1(BM32Input):
    def parse_config(self, config):
        super(EH1, self).parse_config(config)
        self.config.centralpixel = util.parse_tuple(
            config.pop("centralpixel", None), length=2, type=int
        )
        self.config.UB = util.parse_tuple(config.pop("ub", None), length=9, type=float)

    def process_image(self, scanparams, pointparams, edf):
        delta, omega, alfa, beta, chi, phi, mon, transm = pointparams
        wavelength, UB = scanparams

        image = edf.GetData(0)
        header = edf.GetHeader(0)

        weights = numpy.ones_like(image)

        if not self.config.centralpixel:
            self.config.centralpixel = (int(header["y_beam"]), int(header["x_beam"]))
        if not self.config.sdd:
            self.config.sdd = float(header["det_sample_dist"])

        if self.config.background:
            data = image / mon
        else:
            data = image / mon / transm

        if mon == 0:
            raise errors.BackendError(
                "Monitor is zero, this results in empty output. Scannumber = {0}, pointnumber = {1}. Did you forget to open the shutter?".format(
                    self.dbg_scanno, self.dbg_pointno
                )
            )

        util.status(
            "{4}| beta: {0:.3f}, delta: {1:.3f}, omega: {2:.3f}, alfa: {3:.3f}".format(
                beta, delta, omega, alfa, time.ctime(time.time())
            )
        )

        # pixels to angles
        pixelsize = numpy.array(self.config.pixelsize)
        sdd = self.config.sdd

        app = numpy.arctan(pixelsize / sdd) * 180 / numpy.pi

        centralpixel = self.config.centralpixel  # (column, row) = (delta, gamma)
        beta_range = -app[1] * (numpy.arange(data.shape[1]) - centralpixel[1]) + beta
        delta_range = app[0] * (numpy.arange(data.shape[0]) - centralpixel[0]) + delta

        # masking
        if self.config.maskmatrix is not None:
            if self.config.maskmatrix.shape != data.shape:
                raise errors.BackendError(
                    "The mask matrix does not have the same shape as the images"
                )
            weights *= self.config.maskmatrix

        delta_range = delta_range[self.config.ymask]
        beta_range = beta_range[self.config.xmask]

        weights = self.apply_mask(weights, self.config.xmask, self.config.ymask)
        intensity = self.apply_mask(data, self.config.xmask, self.config.ymask)

        intensity = numpy.rot90(intensity)
        intensity = numpy.fliplr(intensity)
        intensity = numpy.flipud(intensity)

        weights = numpy.rot90(weights)
        weights = numpy.fliplr(weights)
        weights = numpy.flipud(weights)

        # polarisation correction
        delta_grid, beta_grid = numpy.meshgrid(delta_range, beta_range)
        Pver = (
            1
            - numpy.sin(delta_grid * numpy.pi / 180.0) ** 2
            * numpy.cos(beta_grid * numpy.pi / 180.0) ** 2
        )
        # intensity /= Pver

        return (
            intensity,
            weights,
            (wavelength, UB, beta_range, delta_range, omega, alfa, chi, phi),
        )

    def get_point_params(self, scan, first, last):
        sl = slice(first, last + 1)

        DEL, OME, ALF, BET, CHI, PHI, MON, TRANSM = list(range(8))
        params = numpy.zeros(
            (last - first + 1, 8)
        )  # gamma delta theta chi phi mu mon transm
        params[:, CHI] = 0  # scan.motorpos('CHI')
        params[:, PHI] = 0  # scan.motorpos('PHI')

        params[:, OME] = scan.datacol("omecnt")[sl]
        params[:, BET] = scan.datacol("betcnt")[sl]
        params[:, DEL] = scan.datacol("delcnt")[sl]
        params[:, MON] = scan.datacol("Monitor")[sl]

        # params[:, TRANSM] = scan.datacol('transm')[sl]
        params[:, TRANSM] = 1

        params[:, ALF] = scan.datacol("alfcnt")[sl]
        return params


def load_matrix(filename):
    if filename == None:
        return None
    if os.path.exists(filename):
        ext = os.path.splitext(filename)[-1]
        if ext == ".txt":
            return numpy.array(numpy.loadtxt(filename), dtype=numpy.bool)
        elif ext == ".npy":
            return numpy.array(numpy.load(filename), dtype=numpy.bool)
        elif ext == ".edf":
            return numpy.array(EdfFile.EdfFile(filename).getData(0), dtype=numpy.bool)
        else:
            raise ValueError(
                "unknown extension {0}, unable to load matrix!\n".format(ext)
            )
    else:
        raise IOError(
            "filename: {0} does not exist. Can not load matrix".format(filename)
        )



from typing import Any, Dict, Tuple

import os
import time
import itertools
import subprocess
import multiprocessing

from . import util, errors, space


class Destination(object):
    type = filename = overwrite = value = config = limits = None
    opts = {}  # type: Dict[Any, Any]

    def set_final_filename(self, filename, overwrite):
        self.type = "final"
        self.filename = filename
        self.overwrite = overwrite

    def set_final_options(self, opts):
        if opts is not False:
            self.opts = opts

    def set_limits(self, limits):
        self.limits = limits

    def set_config(self, conf):
        self.config = conf

    def set_tmp_filename(self, filename):
        self.type = "tmp"
        self.filename = filename

    def set_memory(self):
        self.type = "memory"

    def store(self, verse):
        self.value = None
        if verse.dimension == 0:
            raise ValueError("Empty output, Multiverse contains no spaces")
        if self.type == "memory":
            self.value = verse
        elif self.type == "tmp":
            verse.tofile(self.filename)
            # verse.tovti(self.filename + ".vti")
        elif self.type == "final":
            for sp, fn in zip(verse.spaces, self.final_filenames()):
                sp.config = self.config
                sp.tofile(fn)
                # sp.tovti(fn + ".vti")

    def retrieve(self):
        if self.type == "memory":
            return self.value

    def final_filenames(self):
        fns = []
        if self.limits is not None:
            base, ext = os.path.splitext(self.filename)
            for limlabel in util.limit_to_filelabel(self.limits):
                fn = (base + "_" + limlabel + ext).format(**self.opts)
                if not self.overwrite:
                    fn = util.find_unused_filename(fn)
                fns.append(fn)
        else:
            fn = self.filename.format(**self.opts)
            if not self.overwrite:
                fn = util.find_unused_filename(fn)
            fns.append(fn)
        return fns


class DispatcherBase(util.ConfigurableObject):
    def __init__(self, config, main):
        self.main = main
        super(DispatcherBase, self).__init__(config)

    def parse_config(self, config):
        super(DispatcherBase, self).parse_config(config)
        self.config.destination = Destination()
        # optional 'output.hdf5' by default
        destination = config.pop("destination", "output.hdf5")
        # by default: numbered files in the form output_  # .hdf5:
        overwrite = util.parse_bool(config.pop("overwrite", "false"))
        # explicitly parsing the options first helps with the debugging
        self.config.destination.set_final_filename(destination, overwrite)
        # ip adress of the running gui awaiting the spaces
        self.config.host = config.pop("host", None)
        # port of the running gui awaiting the spaces
        self.config.port = config.pop("port", None)
        # previewing the data, if true, also specify host and port
        self.config.send_to_gui = util.parse_bool(config.pop("send_to_gui", "false"))

    # provides the possiblity to send the results to the gui over the network
    def send(self, verses):
        if self.config.send_to_gui or (
            self.config.host is not None and self.config.host is not None
        ):
            # only continue of ip is specified and send_to_server is flagged
            for M in verses:
                if self.config.destination.limits is None:
                    sp = M.spaces[0]
                    if isinstance(sp, space.Space):
                        util.socket_send(
                            self.config.host,
                            int(self.config.port),
                            util.serialize(sp, ",".join(self.main.config.command)),
                        )  # noqa
                else:
                    for sp, label in zip(
                        M.spaces,
                        util.limit_to_filelabel(self.config.destination.limits),
                    ):  # noqa
                        if isinstance(sp, space.Space):
                            util.socket_send(
                                self.config.host,
                                int(self.config.port),
                                util.serialize(
                                    sp,
                                    "{0}_{1}".format(
                                        ",".join(self.main.config.command), label
                                    ),
                                ),
                            )  # noqa
                yield M
        else:
            for M in verses:
                yield M

    def has_specific_task(self):
        return False

    def process_jobs(self, jobs):
        raise NotImplementedError

    def sum(self, results):
        raise NotImplementedError


# The simplest possible dispatcher. Does the work all by itself on a single
# thread/core/node. 'Local' will most likely suit your needs better.
class SingleCore(DispatcherBase):
    def process_jobs(self, jobs):
        for job in jobs:
            yield self.main.process_job(job)

    def sum(self, results):
        return space.chunked_sum(self.send(results))


# Base class for Dispatchers using subprocesses to do some work.
class ReentrantBase(DispatcherBase):
    actions = ("user",)  # type: Tuple[str, ...]

    def parse_config(self, config):
        super(ReentrantBase, self).parse_config(config)
        self.config.action = config.pop("action", "user").lower()
        if self.config.action not in self.actions:
            raise errors.ConfigError(
                "action {0} not recognized for {1}".format(
                    self.config.action, self.__class__.__name__
                )
            )  # noqa

    def has_specific_task(self):
        if self.config.action == "user":
            return False
        else:
            return True

    def run_specific_task(self, command):
        raise NotImplementedError


# Dispatch multiple worker processes locally, while doing the
# summation in the main process
class Local(ReentrantBase):
    # OFFICIAL API
    actions = "user", "job"

    def parse_config(self, config):
        super(Local, self).parse_config(config)
        # optionally, specify number of cores (autodetect by default)
        self.config.ncores = int(config.pop("ncores", 0))
        if self.config.ncores <= 0:
            self.config.ncores = multiprocessing.cpu_count()

    def process_jobs(self, jobs):
        # note: SingleCore will be marginally faster
        pool = multiprocessing.Pool(self.config.ncores)
        map = pool.imap_unordered

        configs = (self.prepare_config(job) for job in jobs)
        for result in map(self.main.get_reentrant(), configs):
            yield result

    def sum(self, results):
        return space.chunked_sum(self.send(results))

    def run_specific_task(self, command):
        if command:
            raise errors.SubprocessError(
                "invalid command, too many parameters: '{0}'".format(command)
            )  # noqa
        if self.config.action == "job":
            result = self.main.process_job(self.config.job)
            self.config.destination.store(result)

    # UTILITY
    def prepare_config(self, job):
        config = self.main.clone_config()
        config.dispatcher.destination.set_memory()
        config.dispatcher.action = "job"
        config.dispatcher.job = job
        return config, ()


# Dispatch many worker processes on an Oar cluster.


class Oar(ReentrantBase):
    # OFFICIAL API
    actions = "user", "process"

    def parse_config(self, config):
        super(Oar, self).parse_config(config)
        # Optional, current directory by default
        self.config.tmpdir = config.pop("tmpdir", os.getcwd())
        # optionally, tweak oarsub parameters
        self.config.oarsub_options = config.pop("oarsub_options", "walltime=0:15")
        # optionally, override default location of python and/or
        # BINoculars installation
        self.config.executable = config.pop(
            "executable", " ".join(util.get_python_executable())
        )  # noqa

    def process_jobs(self, jobs):
        self.configfiles = []
        self.intermediates = []
        clusters = util.cluster_jobs2(jobs, self.main.input.config.target_weight)
        for jobscluster in clusters:
            uniq = util.uniqid()
            jobconfig = os.path.join(
                self.config.tmpdir, "binoculars-{0}-jobcfg.zpi".format(uniq)
            )
            self.configfiles.append(jobconfig)

            config = self.main.clone_config()
            interm = os.path.join(
                self.config.tmpdir, "binoculars-{0}-jobout.hdf5".format(uniq)
            )
            self.intermediates.append(interm)
            config.dispatcher.destination.set_tmp_filename(interm)
            config.dispatcher.sum = ()

            config.dispatcher.action = "process"
            config.dispatcher.jobs = jobscluster
            util.zpi_save(config, jobconfig)
            yield self.oarsub(jobconfig)

        # if all jobs are sent to the cluster send the process that
        # sums all other jobs
        uniq = util.uniqid()
        jobconfig = os.path.join(
            self.config.tmpdir, "binoculars-{0}-jobcfg.zpi".format(uniq)
        )
        self.configfiles.append(jobconfig)
        config = self.main.clone_config()
        config.dispatcher.sum = self.intermediates
        config.dispatcher.action = "process"
        config.dispatcher.jobs = ()
        util.zpi_save(config, jobconfig)
        yield self.oarsub(jobconfig)

    def sum(self, results):
        jobs = list(results)
        jobscopy = jobs[:]
        self.oarwait(jobs)
        self.oar_cleanup(jobscopy)
        return True

    def run_specific_task(self, command):
        if (
            self.config.action != "process"
            or (not self.config.jobs and not self.config.sum)
            or command
        ):
            raise errors.SubprocessError(
                "invalid command, too many parameters or no jobs/sum given"
            )  # noqa

        jobs = sum = space.EmptyVerse()
        if self.config.jobs:
            jobs = space.verse_sum(
                self.send(self.main.process_job(job) for job in self.config.jobs)
            )
        if self.config.sum:
            sum = space.chunked_sum(
                space.Multiverse.fromfile(src)
                for src in util.yield_when_exists(self.config.sum)
            )  # noqa
        self.config.destination.store(jobs + sum)

    # calling OAR
    @staticmethod
    def subprocess_run(*command):
        process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        output, unused_err = process.communicate()
        retcode = process.poll()
        return retcode, output

    def oarsub(self, *args):
        command = "{0} process {1}".format(
            self.config.executable, " ".join(args)
        )  # noqa
        ret, output = self.subprocess_run(
            "oarsub", "-l {0}".format(self.config.oarsub_options), command
        )  # noqa
        if ret == 0:
            lines = output.split("\n")
            for line in lines:
                if line.startswith("OAR_JOB_ID="):
                    void, jobid = line.split("=")
                    util.status(
                        "{0}: Launched job {1}".format(time.ctime(), jobid)
                    )  # noqa
                    return jobid.strip()
        return False

    def oarstat(self, jobid):
        # % oarstat -s -j 5651374
        # 5651374: Running
        # % oarstat -s -j 5651374
        # 5651374: Finishing
        ret, output = self.subprocess_run("oarstat", "-s", "-j", str(jobid))
        if ret == 0:
            for n in output.split("\n"):
                if n.startswith(str(jobid)):
                    job, status = n.split(":")
            return status.strip()
        else:
            return "Unknown"

    def oarwait(self, jobs, remaining=0):
        if len(jobs) > remaining:
            util.status(
                "{0}: getting status of {1} jobs...".format(time.ctime(), len(jobs))
            )  # noqa
        else:
            return

        delay = util.loop_delayer(30)
        while len(jobs) > remaining:
            next(delay)
            i = 0
            R = 0
            W = 0
            U = 0

            while i < len(jobs):
                state = self.oarstat(jobs[i])
                if state == "Running":
                    R += 1
                elif state in ("Waiting", "toLaunch", "Launching"):
                    W += 1
                elif state == "Unknown":
                    U += 1
                else:  # assume state == 'Finishing' or 'Terminated'
                    # but don't wait on something unknown  # noqa
                    del jobs[i]
                    i -= 1  # otherwise it skips a job
                i += 1
            util.status(
                "{0}: {1} jobs to go. {2} waiting, {3} running, {4} unknown.".format(
                    time.ctime(), len(jobs), W, R, U
                )
            )  # noqa
        util.statuseol()

    def oar_cleanup(self, jobs):
        # cleanup:
        for f in itertools.chain(self.configfiles, self.intermediates):
            try:
                os.remove(f)
            except Exception as e:
                print("unable to remove {0}: {1}".format(f, e))

        errorfn = []

        for jobid in jobs:
            errorfilename = "OAR.{0}.stderr".format(jobid)

            if os.path.exists(errorfilename):
                with open(errorfilename, "r") as fp:
                    errormsg = fp.read()
                if len(errormsg) > 0:
                    errorfn.append(errorfilename)
                    print(
                        "Critical error: OAR Job {0} failed with the following error: \n{1}".format(
                            jobid, errormsg
                        )
                    )  # noqa

        if len(errorfn) > 0:
            print(
                "Warning! {0} job(s) failed. See above for the details or the error log files: {1}".format(
                    len(errorfn), ", ".join(errorfn)
                )
            )  # noqa



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 21 17:02:23 2020

@author: Prevot
"""
from collections import OrderedDict
from copy import deepcopy
import numpy as np
from scipy.integrate import dblquad
from scipy.special import erf
from lmfit import lineshapes
from lmfit import Model as LmfitModel
from lmfit.models import COMMON_INIT_DOC, COMMON_GUESS_DOC, update_param_vals
import operator


def linear2D(x, slope_x=0.0, slope_y=0.0, bg=0.0):
    return bg + x[0] * slope_x + x[1] * slope_y


"""*****************************************************************************************************"""


def lorenzianpower2D(
    x, amp=1.0, loc0=0.0, loc1=0.0, width0=1.0, width1=1.0, th=0.0, power=2.0
):
    cth = np.cos(th)
    sth = np.sin(th)
    x0 = x[0] - loc0
    x1 = x[1] - loc1
    x0, x1 = x0 * cth - x1 * sth, x0 * sth + x1 * cth  # rotation
    if power == 2.0:
        return amp / (1.0 + ((x0 / width0) ** 2 + ((x1 / width1) ** 2)))
    else:
        return amp / (
            1.0 + ((x0 / width0) ** 2 + ((x1 / width1) ** 2)) ** (power / 2.0)
        )


def lorenzianpower(
    y, x, amp=1.0, loc0=0.0, loc1=0.0, width0=1.0, width1=1.0, th=0.0, power=2.0
):
    cth = np.cos(th)
    sth = np.sin(th)
    x0 = x - loc0
    x1 = y - loc1
    x0, x1 = x0 * cth - x1 * sth, x0 * sth + x1 * cth  # rotation
    if power == 2.0:
        return amp / (1.0 + ((x0 / width0) ** 2 + ((x1 / width1) ** 2)))
    else:
        return amp / (
            1.0 + ((x0 / width0) ** 2 + ((x1 / width1) ** 2)) ** (power / 2.0)
        )


def lorenzianpowerint(
    amp=1.0, loc0=0.0, loc1=0.0, width0=1.0, width1=1.0, th=0.0, power=2.0
):
    return dblquad(
        lambda y, x: lorenzianpoly(y, x, amp, loc0, loc1, width0, width1, th, power),
        -np.inf,
        np.inf,
        lambda x: -np.inf,
        lambda x: np.inf,
    )


"""*****************************************************************************************************"""


def lorenziangauss2D(x, amp=1.0, loc0=0.0, loc1=0.0, width0=1.0, width1=1.0, th=0.0):
    cth = np.cos(th)
    sth = np.sin(th)
    x0 = x[0] - loc0
    x1 = x[1] - loc1
    x0, x1 = x0 * cth - x1 * sth, x0 * sth + x1 * cth  # rotation
    return amp * np.exp(-((x1 / width1) ** 2)) / (1.0 + ((x0 / width0) ** 2))


def lorenziangaussint(amp=1.0, loc0=0.0, loc1=0.0, width0=1.0, width1=1.0, th=0.0):
    return (amp * np.pi * np.sqrt(np.pi) * width0 * width1, 0.0)


"""*****************************************************************************************************"""


def lorenzianpoly2D(
    x, amp=1.0, loc0=0.0, loc1=0.0, width0=1.0, width1=1.0, eta0=1.0, eta1=1.0, th=0.0
):
    # eta is the ratio of lorenzian component, if eta=0, it is lorenzian, if eta=1 it is like 1/(1+x**4)
    cth = np.cos(th)
    sth = np.sin(th)
    x0 = x[0] - loc0
    x1 = x[1] - loc1
    x0, x1 = x0 * cth - x1 * sth, x0 * sth + x1 * cth  # rotation
    u0 = (x0 / width0) ** 2
    u1 = (x1 / width1) ** 2
    return amp / (1.0 + (1.0 - eta0 + eta0 * u0) * u0 + (1.0 - eta1 + eta1 * u1) * u1)


def lorenzianpoly(
    y,
    x,
    amp=1.0,
    loc0=0.0,
    loc1=0.0,
    width0=1.0,
    width1=1.0,
    eta0=1.0,
    eta1=1.0,
    th=0.0,
):
    # same with x y coordinates
    cth = np.cos(th)
    sth = np.sin(th)
    x0 = x - loc0
    y0 = y - loc1
    x0, y0 = x0 * cth - y0 * sth, x0 * sth + y0 * cth  # rotation
    u0 = (x0 / width0) ** 2
    u1 = (y0 / width1) ** 2
    return amp / (1.0 + (1.0 - eta0 + eta0 * u0) * u0 + (1.0 - eta1 + eta1 * u1) * u1)


def lorenzianpolyint(
    amp=1.0, loc0=0.0, loc1=0.0, width0=1.0, width1=1.0, eta0=1.0, eta1=1.0, th=0.0
):
    return dblquad(
        lambda y, x: lorenzianpoly(
            y, x, amp, loc0, loc1, width0, width1, eta0, eta1, th
        ),
        -np.inf,
        np.inf,
        lambda x: -np.inf,
        lambda x: np.inf,
    )


"""*****************************************************************************************************"""


def lorenziandoor2D(
    x, amp=1.0, loc0=0.0, loc1=0.0, width0=1.0, width1=1.0, eta=1.0, th=0.0
):
    cth = np.cos(th)
    sth = np.sin(th)
    x0 = x[0] - loc0
    x1 = x[1] - loc1
    x0, x1 = x0 * cth - x1 * sth, x0 * sth + x1 * cth  # rotation
    return (
        amp
        / (1.0 + ((x0 / width0) ** 2))
        * (
            erf((x1 + width1 / 2.0) / (eta * width1))
            - erf((x1 - width1 / 2.0) / (eta * width1))
        )
        / 2.0
    )


def lorenziandoorint(
    amp=1.0, loc0=0.0, loc1=0.0, width0=1.0, width1=1.0, eta=1.0, th=0.0
):
    return (amp * np.pi * width0 * width1, 0.0)


"""*****************************************************************************************************"""


def nointegration(*args, **kwargs):
    return (0.0, 0.0)


# we jsut add the possibility to compute integrals for modelss
class Model(LmfitModel):
    def __init__(
        self,
        func,
        independent_vars=None,
        param_names=None,
        nan_policy="raise",
        missing=None,
        prefix="",
        name=None,
        **kws
    ):
        self.intfunc = nointegration
        super(Model, self).__init__(
            func=func,
            independent_vars=independent_vars,
            param_names=param_names,
            nan_policy=nan_policy,
            missing=missing,
            prefix=prefix,
            name=name,
            **kws
        )

    def __add__(self, other):
        """+"""
        return CompositeModel(self, other, operator.add)

    def __sub__(self, other):
        """-"""
        return CompositeModel(self, other, operator.sub)

    def __mul__(self, other):
        """*"""
        return CompositeModel(self, other, operator.mul)

    def __div__(self, other):
        """/"""
        return CompositeModel(self, other, operator.truediv)

    def __truediv__(self, other):
        """/"""
        return CompositeModel(self, other, operator.truediv)

    def eval_integral(self, params=None, **kwargs):
        # evaluate the integral of the peak, and associated scipy.integrate error (not the fit uncertainty)
        # needs to define self.intfunc first!
        return self.intfunc(**self.make_funcargs(params, kwargs))

    def eval_component_integrals(self, params=None, **kwargs):
        """Evaluate the model integrals with the supplied parameters.

        Parameters
        -----------
        params : Parameters, optional
            Parameters to use in Model.
        **kwargs : optional
            Additional keyword arguments to pass to model function.

        Returns
        -------
        OrderedDict
            Keys are prefixes for component model, values are value of each
            component.

        """
        key = self._prefix
        if len(key) < 1:
            key = self._name
        return {
            key: self.eval_integral(params=params, **kwargs)
        }  # will be used to update the dictionnary of a composite model


class CompositeModel(Model):
    """Combine two models (`left` and `right`) with a binary operator (`op`)
    into a CompositeModel.

    Normally, one does not have to explicitly create a `CompositeModel`,
    but can use normal Python operators `+`, '-', `*`, and `/` to combine
    components as in::

    >>> mod = Model(fcn1) + Model(fcn2) * Model(fcn3)

    """

    _names_collide = (
        "\nTwo models have parameters named '{clash}'. " "Use distinct names."
    )
    _bad_arg = "CompositeModel: argument {arg} is not a Model"
    _bad_op = "CompositeModel: operator {op} is not callable"
    _known_ops = {
        operator.add: "+",
        operator.sub: "-",
        operator.mul: "*",
        operator.truediv: "/",
    }

    def __init__(self, left, right, op, **kws):
        """
        Parameters
        ----------
        left : Model
            Left-hand model.
        right : Model
            Right-hand model.
        op : callable binary operator
            Operator to combine `left` and `right` models.
        **kws : optional
            Additional keywords are passed to `Model` when creating this
            new model.

        Notes
        -----
        1. The two models must use the same independent variable.

        """
        if not isinstance(left, Model):
            raise ValueError(self._bad_arg.format(arg=left))
        if not isinstance(right, Model):
            raise ValueError(self._bad_arg.format(arg=right))
        if not callable(op):
            raise ValueError(self._bad_op.format(op=op))

        self.left = left
        self.right = right
        self.op = op

        name_collisions = set(left.param_names) & set(right.param_names)
        if len(name_collisions) > 0:
            msg = ""
            for collision in name_collisions:
                msg += self._names_collide.format(clash=collision)
            raise NameError(msg)

        # we assume that all the sub-models have the same independent vars
        if "independent_vars" not in kws:
            kws["independent_vars"] = self.left.independent_vars
        if "nan_policy" not in kws:
            kws["nan_policy"] = self.left.nan_policy

        def _tmp(self, *args, **kws):
            pass

        Model.__init__(self, _tmp, **kws)

        for side in (left, right):
            prefix = side.prefix
            for basename, hint in side.param_hints.items():
                self.param_hints["%s%s" % (prefix, basename)] = hint

    def _parse_params(self):
        self._func_haskeywords = (
            self.left._func_haskeywords or self.right._func_haskeywords
        )
        self._func_allargs = self.left._func_allargs + self.right._func_allargs
        self.def_vals = deepcopy(self.right.def_vals)
        self.def_vals.update(self.left.def_vals)
        self.opts = deepcopy(self.right.opts)
        self.opts.update(self.left.opts)

    def _reprstring(self, long=False):
        return "(%s %s %s)" % (
            self.left._reprstring(long=long),
            self._known_ops.get(self.op, self.op),
            self.right._reprstring(long=long),
        )

    def eval(self, params=None, **kwargs):
        """TODO: docstring in public method."""
        return self.op(
            self.left.eval(params=params, **kwargs),
            self.right.eval(params=params, **kwargs),
        )

    def eval_components(self, **kwargs):
        """Return OrderedDict of name, results for each component."""
        out = OrderedDict(self.left.eval_components(**kwargs))
        out.update(self.right.eval_components(**kwargs))
        return out

    def eval_component_integrals(self, params=None, **kwargs):
        """Return OrderedDict of name, results for each component."""
        out = OrderedDict(self.left.eval_component_integrals(params, **kwargs))
        out.update(self.right.eval_component_integrals(params, **kwargs))
        return out

    @property
    def param_names(self):
        """Return parameter names for composite model."""
        return self.left.param_names + self.right.param_names

    @property
    def components(self):
        """Return components for composite model."""
        return self.left.components + self.right.components

    def _get_state(self):
        return (self.left._get_state(), self.right._get_state(), self.op.__name__)

    def _set_state(self, state, funcdefs=None):
        return _buildmodel(state, funcdefs=funcdefs)

    def _make_all_args(self, params=None, **kwargs):
        """Generate **all** function arguments for all functions."""
        out = self.right._make_all_args(params=params, **kwargs)
        out.update(self.left._make_all_args(params=params, **kwargs))
        return out


def _buildmodel(state, funcdefs=None):
    """Build model from saved state.

    Intended for internal use only.

    """
    if len(state) != 3:
        raise ValueError("Cannot restore Model")
    known_funcs = {}
    for fname in lineshapes.functions:
        fcn = getattr(lineshapes, fname, None)
        if callable(fcn):
            known_funcs[fname] = fcn
    if funcdefs is not None:
        known_funcs.update(funcdefs)

    left, right, op = state
    if op is None and right is None:
        (fname, fcndef, name, prefix, ivars, pnames, phints, nan_policy, opts) = left
        if not callable(fcndef) and fname in known_funcs:
            fcndef = known_funcs[fname]

        if fcndef is None:
            raise ValueError("Cannot restore Model: model function not found")

        model = Model(
            fcndef,
            name=name,
            prefix=prefix,
            independent_vars=ivars,
            param_names=pnames,
            nan_policy=nan_policy,
            **opts
        )

        for name, hint in phints.items():
            model.set_param_hint(name, **hint)
        return model
    else:
        lmodel = _buildmodel(left, funcdefs=funcdefs)
        rmodel = _buildmodel(right, funcdefs=funcdefs)
        return CompositeModel(lmodel, rmodel, getattr(operator, op))


# Models derived from the lmfit Model class
# Need to give param_names if we want to keep the correct order


class LorenzianPower2DModel(Model):
    """Lorentzian model with variable power, with seven Parameters.

    Defined as:

    f(x; amp, loc0, loc1, width0, width1, th) = amp/(1.+((x0/width0)**2+(x1/width1)**2)**(power/2.)))

    """

    EVALUATE = "peak"
    ICON = "lorenzpower.jpg"
    NAME = "Lorenzian with extra power"

    def __init__(self, independent_vars=["x"], prefix="", nan_policy="raise", **kwargs):
        kwargs.update(
            {
                "prefix": prefix,
                "nan_policy": nan_policy,
                "independent_vars": independent_vars,
            }
        )
        super(LorenzianPower2DModel, self).__init__(
            lorenzianpower2D,
            param_names=["amp", "loc0", "loc1", "width0", "width1", "th", "power"],
            **kwargs
        )
        self.set_param_hint("amp", value=1.0)
        self.set_param_hint("loc0", value=0.0)
        self.set_param_hint("loc1", value=0.0)
        self.set_param_hint("width0", value=1.0, min=0)
        self.set_param_hint("width1", value=1.0, min=0)
        self.set_param_hint("th", value=0.0, vary=False)
        self.set_param_hint("power", value=2.0, vary=False)
        self.intfunc = lorenzianpowerint

    def guess(self, data, x=None, **kwargs):
        """Estimate initial model parameter values from data."""
        amp, loc0, loc1, width0, width1, th, power = 1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 2.0

        if x is not None:
            n = np.argmax(data)
            amp = data[n]
            loc0 = x[0][n]
            loc1 = x[1][n]
            width0 = (np.amax(x) - np.amin(x)) / 2.0
            width1 = (np.amax(x) - np.amin(x)) / 2.0
            if width0 == 0:
                width0 = 1.0
            if width1 == 0:
                width1 = 1.0

        pars = self.make_params(
            amp=amp,
            loc0=loc0,
            loc1=loc1,
            width0=width0,
            width1=width1,
            th=th,
            power=power,
        )
        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.__doc__ = COMMON_INIT_DOC
    guess.__doc__ = COMMON_GUESS_DOC


class LorenzianPoly2DModel(Model):
    """Lorentzian model with polynomial, with seven Parameters.

    Defined as:

    f(x; amp, loc0, loc1, width0, width1, th) = amp/(1.+((x0/width0)**2+(x1/width1)**2)**(power/2.)))

    """

    EVALUATE = "peak"
    ICON = "lorenzpoly.jpg"
    NAME = "Lorenzian with x**4 term"

    def __init__(self, independent_vars=["x"], prefix="", nan_policy="raise", **kwargs):
        kwargs.update(
            {
                "prefix": prefix,
                "nan_policy": nan_policy,
                "independent_vars": independent_vars,
            }
        )
        super(LorenzianPoly2DModel, self).__init__(
            lorenzianpoly2D,
            param_names=[
                "amp",
                "loc0",
                "loc1",
                "width0",
                "width1",
                "eta0",
                "eta1",
                "th",
            ],
            **kwargs
        )
        self.set_param_hint("amp", value=1.0)
        self.set_param_hint("loc0", value=0.0)
        self.set_param_hint("loc1", value=0.0)
        self.set_param_hint("width0", value=1.0, min=0)
        self.set_param_hint("width1", value=1.0, min=0)
        self.set_param_hint("eta0", value=0.5, min=0, max=1)
        self.set_param_hint("eta1", value=0.5, min=0, max=1)
        self.set_param_hint("th", value=0.0, min=-np.pi, max=np.pi)
        self.intfunc = lorenzianpolyint

    def guess(self, data, x=None, **kwargs):
        """Estimate initial model parameter values from data."""
        amp, loc0, loc1, width0, width1, th, power = 1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 2.0

        if x is not None:
            n = np.argmax(data)
            amp = data[n]
            loc0 = x[0][n]
            loc1 = x[1][n]
            width0 = (np.amax(x) - np.amin(x)) / 2.0
            width1 = (np.amax(x) - np.amin(x)) / 2.0
            if width0 == 0:
                width0 = 1.0
            if width1 == 0:
                width1 = 1.0

        pars = self.make_params(
            amp=amp,
            loc0=loc0,
            loc1=loc1,
            width0=width0,
            width1=width1,
            th=th,
            power=power,
        )
        return update_param_vals(pars, self.prefix, **kwargs)

    def eval_integral(self, params=None, **kwargs):
        """Evaluate the integral of the model with supplied parameters and keyword arguments.

        Parameters
        -----------
        params : Parameters, optional
            Parameters to use in Model.
        **kwargs : optional
            Additional keyword arguments to pass to model function.

        Returns
        -------
        numpy.ndarray
            Value of model given the parameters and other arguments.

        Notes
        -----
        1. if `params` is None, the values for all parameters are
        expected to be provided as keyword arguments.  If `params` is
        given, and a keyword argument for a parameter value is also given,
        the keyword argument will be used.

        2. all non-parameter arguments for the model function, **including
        all the independent variables** will need to be passed in using
        keyword arguments.

        """
        return self.intfunc(**self.make_funcargs(params, kwargs))

    __init__.__doc__ = COMMON_INIT_DOC
    guess.__doc__ = COMMON_GUESS_DOC


class LorenzianDoorModel(Model):
    """Lorentzian in one direction, door (with 2 erf) in the other direction, with seven Parameters.

    Defined as:

    f(x; amp, loc0, loc1, width0, width1, eta, th) = (amp/(1.+((x0/width0)**2)*(erf((y0-width0/2)/eta)-erf((y0+width0/2)))

    """

    EVALUATE = "peak"
    ICON = "lorenzgauss.jpg"
    NAME = "Lorenzian/door"

    def __init__(self, independent_vars=["x"], prefix="", nan_policy="raise", **kwargs):
        kwargs.update(
            {
                "prefix": prefix,
                "nan_policy": nan_policy,
                "independent_vars": independent_vars,
            }
        )
        super(LorenzianDoorModel, self).__init__(
            lorenziandoor2D,
            param_names=["amp", "loc0", "loc1", "width0", "width1", "eta", "th"],
            **kwargs
        )
        self.set_param_hint("amp", value=1.0)
        self.set_param_hint("loc0", value=0.0)
        self.set_param_hint("loc1", value=0.0)
        self.set_param_hint("width0", value=1.0, min=0)
        self.set_param_hint("width1", value=1.0, min=0)
        self.set_param_hint("eta", value=0.2, min=0)
        self.set_param_hint("th", value=0.0, min=-np.pi, max=np.pi)
        self.intfunc = lorenziandoorint

    def guess(self, data, x=None, **kwargs):
        """Estimate initial model parameter values from data."""
        amp, loc0, loc1, width0, width1, eta, th = 1.0, 0.0, 0.0, 1.0, 1.0, 1.0, 0.0

        if x is not None:
            n = np.argmax(data)
            amp = data[n]
            loc0 = x[0][n]
            loc1 = x[1][n]
            width0 = (np.amax(x) - np.amin(x)) / 2.0
            width1 = (np.amax(x) - np.amin(x)) / 2.0
            if width0 == 0:
                width0 = 1.0
            if width1 == 0:
                width1 = 1.0

        pars = self.make_params(
            amp=amp, loc0=loc0, loc1=loc1, width0=width0, width1=width1, eta=eta, th=th
        )
        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.__doc__ = COMMON_INIT_DOC
    guess.__doc__ = COMMON_GUESS_DOC


class LorenzianGaussModel(Model):
    """Lorentzian in one direction, gaussian in the other direction, with six Parameters.

    Defined as:

    f(x; amp, loc0, loc1, width0, width1, th) = (amp/(1.+((x0/width0)**2)*np.exp(-(x1/width1)**2)))

    """

    EVALUATE = "peak"
    ICON = "lorenzgauss.jpg"
    NAME = "Lorenzian/gaussian"

    def __init__(self, independent_vars=["x"], prefix="", nan_policy="raise", **kwargs):
        kwargs.update(
            {
                "prefix": prefix,
                "nan_policy": nan_policy,
                "independent_vars": independent_vars,
            }
        )
        super(LorenzianGaussModel, self).__init__(
            lorenziangauss2D,
            param_names=["amp", "loc0", "loc1", "width0", "width1", "th"],
            **kwargs
        )
        self.set_param_hint("amp", value=1.0)
        self.set_param_hint("loc0", value=0.0)
        self.set_param_hint("loc1", value=0.0)
        self.set_param_hint("width0", value=1.0, min=0)
        self.set_param_hint("width1", value=1.0, min=0)
        self.set_param_hint("th", value=0.0, min=-np.pi, max=np.pi)
        self.intfunc = lorenziangaussint

    def guess(self, data, x=None, **kwargs):
        """Estimate initial model parameter values from data."""
        amp, loc0, loc1, width0, width1, th = 1.0, 0.0, 0.0, 1.0, 1.0, 0.0

        if x is not None:
            n = np.argmax(data)
            amp = data[n]
            loc0 = x[0][n]
            loc1 = x[1][n]
            width0 = (np.amax(x) - np.amin(x)) / 2.0
            width1 = (np.amax(x) - np.amin(x)) / 2.0
            if width0 == 0:
                width0 = 1.0
            if width1 == 0:
                width1 = 1.0

        pars = self.make_params(
            amp=amp, loc0=loc0, loc1=loc1, width0=width0, width1=width1, th=th
        )
        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.__doc__ = COMMON_INIT_DOC
    guess.__doc__ = COMMON_GUESS_DOC


class Linear2DModel(Model):
    """Lorentzian model, with six Parameters.

    Defined as:

    f(x; slope_x,slope_y,bg) = amp/(1.+(x0/width0)**2+(x1/width1)**2)

    """

    EVALUATE = "points"
    ICON = "planar.jpg"
    NAME = "Linear background"

    def __init__(self, independent_vars=["x"], prefix="", nan_policy="raise", **kwargs):
        kwargs.update(
            {
                "prefix": prefix,
                "nan_policy": nan_policy,
                "independent_vars": independent_vars,
            }
        )
        super(Linear2DModel, self).__init__(
            linear2D, param_names=["bg", "slope_x", "slope_y"], **kwargs
        )
        self.set_param_hint("bg", value=1.0)
        self.set_param_hint("slope_x", value=0.0)
        self.set_param_hint("slope_y", value=0.0)

    def guess(self, data, x=None, **kwargs):
        """Estimate initial model parameter values from data."""
        slope_x, slope_y, bg = 0.0, 0.0, 0.0

        if x is not None:
            sx2 = np.sum(x[0] * x[0])
            sy2 = np.sum(x[1] * x[1])
            sxy = np.sum(x[0] * x[1])
            sx = np.sum(x[0])
            sy = np.sum(x[0])
            szx = np.sum(x[0] * data)
            szy = np.sum(x[0] * data)
            sz = np.sum(data)
            npts = x[0].shape[0]
            a = np.array([[sx2, sxy, sx], [sx, sy2, sy], [sx, sy, npts]])
            b = np.array([szx, szy, sz])
            res = np.linalg.solve(a, b)
            slope_x = res[0]
            slope_y = res[1]
            bg = res[2]
        pars = self.make_params(slope_x=slope_x, slope_y=slope_y, bg=bg)
        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.__doc__ = COMMON_INIT_DOC
    guess.__doc__ = COMMON_GUESS_DOC


list_of_models = []
list_of_2D_models = [
    Linear2DModel,
    LorenzianGaussModel,
    LorenzianPoly2DModel,
    LorenzianPower2DModel,
    LorenzianDoorModel,
]



"""This file is part of the binoculars project.

  The BINoculars library is free software: you can redistribute it
  and/or modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation, either version 3 of
  the License, or (at your option) any later version.

  The BINoculars library is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with the hkl library.  If not, see
  <http://www.gnu.org/licenses/>.

  Copyright (C) 2015-2019 Synchrotron SOLEIL
                          L'Orme des Merisiers Saint-Aubin
                          BP 48 91192 GIF-sur-YVETTE CEDEX

  Copyright (C) 2012-2015 European Synchrotron Radiation Facility
                          Grenoble, France

  Authors: Willem Onderwaater <onderwaa@esrf.fr>
           Picca Frédéric-Emmanuel <picca@synchrotron-soleil.fr>

"""
from typing import Dict, NamedTuple, Optional, Tuple

import numpy
import math
import os

from gi.repository import Hkl
from h5py import Dataset, File
from numpy import ndarray
from pyFAI.detectors import ALL_DETECTORS

from .soleil import DatasetPathContains, DatasetPathWithAttribute, HItem, get_dataset

# put hkl_matrix_to_numpy into soleil
from .sixs import DataFrame, M, PDataFrame, SIXS, WRONG_ATTENUATION, hkl_matrix_to_numpy
from ..util import status

##################
# Input Backends #
##################


class Values(NamedTuple):
    image: ndarray
    attenuation: Optional[float]
    timestamp: float
    values: Tuple[float]


class Diffabs(SIXS):
    HPATH = {
        "image": DatasetPathWithAttribute("interpretation", b"image"),
        "mu": DatasetPathWithAttribute("long_name", b"d13-1-cx1/ex/dif.1-mu/position"),
        "komega": DatasetPathWithAttribute(
            "long_name", b"d13-1-cx1/ex/dif.1-komega/position"
        ),
        "kappa": DatasetPathWithAttribute(
            "long_name", b"d13-1-cx1/ex/dif.1-kappa/position"
        ),
        "kphi": DatasetPathWithAttribute(
            "long_name", b"d13-1-cx1/ex/dif.1-kphi/position"
        ),
        "gamma": DatasetPathWithAttribute(
            "long_name", b"d13-1-cx1/ex/dif.1-gamma/position"
        ),
        "delta": DatasetPathWithAttribute(
            "long_name", b"d13-1-cx1/ex/dif.1-delta/position"
        ),
        "attenuation": HItem("attenuation", True),
        "timestamp": DatasetPathContains("sensors_timestamps"),
    }

    def get_pointcount(self, scanno: int) -> int:
        # just open the file in order to extract the number of step
        with File(self.get_filename(scanno), "r") as scan:
            path = self.HPATH["image"]
            return get_dataset(scan, path).shape[0]

    def get_attenuation(
        self, index: int, h5_nodes: Dict[str, Dataset], offset: int
    ) -> Optional[float]:
        attenuation = None
        if self.config.attenuation_coefficient is not None:
            try:
                try:
                    node = h5_nodes["attenuation"]
                    if node is not None:
                        attenuation = node[index + offset]
                    else:
                        raise Exception(
                            "you asked for attenuation but the file does not contain attenuation informations."
                        )  # noqa
                except KeyError:
                    attenuation = 1.0
            except ValueError:
                attenuation = WRONG_ATTENUATION
        return attenuation

    def get_timestamp(self, index: int, h5_nodes: Dict[str, Dataset]) -> float:
        timestamp = None
        try:
            timestamp = h5_nodes["timestamp"][index]
        except KeyError:
            timestamp = index
        return timestamp

    def get_values(self, index: int, h5_nodes: Dict[str, Dataset]) -> Values:
        image = h5_nodes["image"][index]
        mu = h5_nodes["mu"][index]
        komega = h5_nodes["komega"][index]
        kappa = h5_nodes["kappa"][index]
        kphi = h5_nodes["kphi"][index]
        gamma = h5_nodes["gamma"][index]
        delta = h5_nodes["delta"][index] + 16.004
        attenuation = self.get_attenuation(index, h5_nodes, index)
        timestamp = self.get_timestamp(index, h5_nodes)

        return Values(
            image, attenuation, timestamp, (mu, komega, kappa, kphi, gamma, delta)
        )

    def process_image(
        self, index: int, dataframe: DataFrame, pixels: ndarray, mask: ndarray
    ) -> Tuple[ndarray, ndarray, Tuple[int, PDataFrame]]:
        status(str(index))

        # extract the data from the h5 nodes
        h5_nodes = dataframe.h5_nodes
        intensity, attenuation, timestamp, values = self.get_values(index, h5_nodes)

        # BEWARE in order to avoid precision problem we convert the
        # uint16 -> float32. (the size of the mantis is on 23 bits)
        # enought to contain the uint16. If one day we use uint32, it
        # should be necessary to convert into float64.
        intensity = intensity.astype("float32")

        weights = None
        if self.config.attenuation_coefficient is not None:
            if attenuation != WRONG_ATTENUATION:
                intensity *= self.config.attenuation_coefficient ** attenuation
                weights = numpy.ones_like(intensity)
                weights *= ~mask
            else:
                weights = numpy.zeros_like(intensity)
        else:
            weights = numpy.ones_like(intensity)
            weights *= ~mask

        k = 2 * math.pi / dataframe.source.wavelength

        hkl_geometry = dataframe.diffractometer.geometry
        hkl_geometry.axis_values_set(values, Hkl.UnitEnum.USER)

        # sample
        hkl_sample = dataframe.sample.sample
        q_sample = hkl_geometry.sample_rotation_get(hkl_sample)
        R = hkl_matrix_to_numpy(q_sample.to_matrix())

        # detector
        hkl_detector = dataframe.detector.detector
        q_detector = hkl_geometry.detector_rotation_get(hkl_detector)
        P = hkl_matrix_to_numpy(q_detector.to_matrix())

        if self.config.detrot is not None:
            P = numpy.dot(P, M(math.radians(self.config.detrot), [1, 0, 0]))

        surface_orientation = self.config.surface_orientation

        pdataframe = PDataFrame(
            pixels,
            k,
            dataframe.diffractometer.ub,
            R,
            P,
            index,
            timestamp,
            surface_orientation,
        )

        return intensity, weights, (index, pdataframe)

    def get_pixels(self, detector):
        # works only for flat detector.
        detector = ALL_DETECTORS[detector.name]()
        y, x, _ = detector.calc_cartesian_positions()
        y0 = y[self.config.centralpixel[1], self.config.centralpixel[0]]
        x0 = x[self.config.centralpixel[1], self.config.centralpixel[0]]
        z = numpy.ones(x.shape) * -1 * self.config.sdd
        # return converted to the hkl library coordinates
        # x -> -y
        # y -> z
        # z -> -x
        return numpy.array([-z, -(x - x0), (y - y0)])

    def get_filename(self, scanno):
        filename = None
        if self.config.nexusdir:
            dirname = self.config.nexusdir
            files = [
                f
                for f in os.listdir(dirname)
                if (
                    (str(scanno).zfill(5) in f)
                    and (os.path.splitext(f)[1] in [".hdf5", ".nxs"])
                )
            ]
            if files is not []:
                filename = os.path.join(dirname, files[0])
        else:
            print(self.config.nexusfile)
            filename = self.config.nexusfile.format(scanno)  # noqa
        if not os.path.exists(filename):
            raise Exception(
                "nexus filename does not exist: {0}".format(filename)
            )  # noqa
        return filename



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 21 14:15:11 2020

@author: Prevot
"""
from guidata.qt.QtCore import SIGNAL

# Emitted when a fit is performed
SIG_FIT_DONE = SIGNAL("fit_done")

# Emitted when a fit is performed
SIG_FIT_SAVED = SIGNAL("fit_saved")

# Emitted by parameter table when try is
SIG_TRY_CHANGED = SIGNAL("try_changed")

SIG_TRIGGERED = SIGNAL("triggered()")

SIG_MODEL_ADDED = SIGNAL("model_added(int)")

SIG_MODEL_REMOVED = SIGNAL("model_removed(int)")

SIG_KEY_PRESSED_EVENT = SIGNAL("keyPressedEvent")



from . import util, errors, dispatcher


class ProjectionBase(util.ConfigurableObject):
    def parse_config(self, config):
        super(ProjectionBase, self).parse_config(config)
        res = config.pop("resolution")  # or just give 1 number for

        # all dimensions Optional, set the limits of the space object
        # in projected coordinates. Syntax is same as numpy
        # e.g. 'limits = [:0,-1:,:], [0:,:-1,:], [:0,:-1,:],
        # [0:,-1:,:]'
        self.config.limits = util.parse_pairs(config.pop("limits", None))
        labels = self.get_axis_labels()
        if self.config.limits is not None:
            for lim in self.config.limits:
                if len(lim) != len(labels):
                    raise errors.ConfigError(
                        "dimension mismatch between projection axes ({0}) and limits specification ({1}) in {2}".format(  # noqa
                            labels, self.config.limits, self.__class__.__name__
                        )
                    )  # noqa
        if "," in res:
            self.config.resolution = util.parse_tuple(res, type=float)
            if not len(labels) == len(self.config.resolution):
                raise errors.ConfigError(
                    "dimension mismatch between projection axes ({0}) and resolution specification ({1}) in {2}".format(  # noqa
                        labels, self.config.resolution, self.__class__.__name__
                    )
                )  # noqa
        else:
            self.config.resolution = tuple([float(res)] * len(labels))

    def project(self, *args):
        raise NotImplementedError

    def get_axis_labels(self):
        raise NotImplementedError


class Job(object):
    weight = 1.0  # estimate of job difficulty (arbitrary units)

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class InputBase(util.ConfigurableObject):
    """Generate and process Job()s.

    Note: there is no guarantee that generate_jobs() and process_jobs() will
    be called on the same instance, not even in the same process or on the
    same computer!"""

    def parse_config(self, config):
        super(InputBase, self).parse_config(config)
        # approximate number of images per job, only useful when
        # running on the oar cluster
        self.config.target_weight = int(config.pop("target_weight", 1000))

    def generate_jobs(self, command):
        """Receives command from user, yields Job() instances"""
        raise NotImplementedError

    def process_job(self, job):
        """Receives a Job() instance, yields (intensity,
args_to_be_sent_to_a_Projection_instance)

        Job()s could have been pickle'd and distributed over a cluster

        """
        self.metadata = util.MetaBase("job", job.__dict__)

    def get_destination_options(self, command):
        """Receives the same command as generate_jobs(), but returns
        dictionary that will be used to .format() the dispatcher:destination
        configuration value."""
        return {}


def get_dispatcher(config, main, default=None):
    return _get_backend(
        config, "dispatcher", dispatcher.DispatcherBase, default=default, args=[main]
    )


def get_input(config, default=None):
    return _get_backend(config, "input", InputBase, default=default)


def get_projection(config, default=None):
    return _get_backend(config, "projection", ProjectionBase, default=default)


def _get_backend(config, section, basecls, default=None, args=[], kwargs={}):
    if isinstance(config, util.ConfigSection):
        return config.class_(config, *args, **kwargs)

    type = config.pop("type", default)
    if type is None:
        raise errors.ConfigError(
            "required option 'type' not given in section '{0}'".format(section)  # noqa
        )
    type = type.strip()

    if ":" in type:
        try:
            modname, clsname = type.split(":")
        except ValueError:
            raise errors.ConfigError(
                "invalid type '{0}' in section '{1}'".format(type, section)
            )
        try:
            backend = __import__(
                "backends.{0}".format(modname), globals(), locals(), [], 1
            )
        except ImportError as e:
            raise errors.ConfigError(
                "unable to import module backends.{0}: {1}".format(modname, e)  # noqa
            )
        module = getattr(backend, modname)
    elif section == "dispatcher":
        module = dispatcher
        clsname = type
    else:
        raise errors.ConfigError(
            "invalid type '{0}' in section '{1}'".format(type, section)
        )

    clsname = clsname.lower()
    names = dict((name.lower(), name) for name in dir(module))
    if clsname in names:
        cls = getattr(module, names[clsname])

        if issubclass(cls, basecls):
            return cls(config, *args, **kwargs)
        else:
            raise errors.ConfigError(
                "type '{0}' not compatible in section '{1}': expected class derived from '{2}', got '{3}'".format(  # noqa
                    type, section, basecls.__name__, cls.__name__
                )
            )  # noqa
    else:
        raise errors.ConfigError(
            "invalid type '{0}' in section '{1}'".format(type, section)
        )  # noqa



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 21 14:57:49 2020

@author: Prevot
"""


def get_xywxwy_cxy(plot, p0, p1):
    # return position and width from a centered circular shape drawn from p0 to p1
    ax, ay = plot.get_axis_id("bottom"), plot.get_axis_id("left")
    x1, y1 = plot.invTransform(ax, p0.x()), plot.invTransform(ay, p0.y())
    x2, y2 = plot.invTransform(ax, p1.x()), plot.invTransform(ay, p1.y())
    return x1, y1, 2.0 * abs(x2 - x1), 2 * abs(y1 - y2)



import sys
import os
import itertools
import numpy

from .. import backend, errors, util


"""
This example backend contains the minimal set of functions needed to construct a backend.
It consists of a child of a backend.InputBase class and a child of a backend.ProjectionBase
class. The backend.Inputbase is collects the data from the measurement. The backend.ProjectionBase
class calculates the new coordinates per pixel.

You can write as much input classes and as much projections in one backend as you prefer, provided that
the output of the inputclass is compatible with projection class. Otherwise you will be
served best by writing a new backend, for the incompatibility will create errors that break the script.
In the configuration file you specify the inputclass and projection needed for the treatment of the dataset.
"""


class QProjection(backend.ProjectionBase):
    def project(self, wavelength, af, delta, omega, ai):
        """
        This class takes as input the tuple of coordinates returned by the process_job 
        method in the backend.InputBase class. Here you specify how to project the coordinates
        that belong to every datapoint. The number of input arguments should match the
        second tuple returned by process_job. The shape of each returned array should match
        the shape of the first argument returned by process_job
        """

        k0 = 2 * numpy.pi / wavelength

        qy = k0 * (numpy.cos(af) * numpy.cos(delta) - numpy.cos(ai) * numpy.cos(omega))
        qx = k0 * (numpy.cos(af) * numpy.sin(delta) - numpy.cos(ai) * numpy.sin(omega))
        qz = k0 * (numpy.sin(af) + numpy.sin(ai))

        return (
            qx.flatten(),
            qy.flatten(),
            qz.flatten(),
        )  # a tuple of numpy.arrays with the same dimension as the number of labels

    def get_axis_labels(self):
        """
        Specify the names of the axes. The number of labels should be equal to the number
        of arrays returned in the project method.
        """
        return "qx", "qy", "qz"


class Input(backend.InputBase):
    def generate_jobs(self, command):
        """
        Command is supplied when the program is started in the terminal. This can used to differentiate between separate datasets
        that will be processed independently.
        """

        scans = util.parse_multi_range(
            ",".join(command).replace(" ", ",")
        )  # parse the command
        for scanno in scans:
            yield backend.Job(scan=scanno)

    def process_job(self, job):
        """
        This methods is a generator that returns the intensity, the weights and a tuple of coordinates that
        will be used for projection. The input is a backend.job object. This objects contains attributes that are supplied
        as keyword arguments in the generate_jobs method when backend.Job is instantiated. You can wet here the weights according
        the behaviour of your detector. To select normal averaging give the weights the value of ones. This array should be the same shape as
        the intensity array.

        This example backend simulates a random path through angular space starting at the origin.
        an example image will be generated using a three dimensional 10-slit interference function.
        The angles are with respect to the sample where af and delta are the angular coordinates
        of the pixels and ai and omega are the in plane and out of plane angles of the incoming beam.
        """
        super(Input, self).process_job(job)  # call super to fix metadeta handling
        scan = job.scan

        # reflects a scan with 100 datapoints
        aaf = numpy.linspace(0, numpy.random.random() * 20, 100)
        adelta = numpy.linspace(0, numpy.random.random() * 20, 100)
        aai = numpy.linspace(0, numpy.random.random() * 20, 100)
        aomega = numpy.linspace(0, numpy.random.random() * 20, 100)
        for af, delta, ai, omega in zip(aaf, adelta, aai, aomega):
            print(
                "af: {0}, delta: {1}, ai: {2}, omega: {3}".format(af, delta, ai, omega)
            )

            # caculating the angles per pixel. The values specified in the configuration file
            # can be used for calculating these values
            pixelsize = numpy.array(self.config.pixelsize)
            sdd = self.config.sdd
            app = numpy.arctan(pixelsize / sdd) * 180 / numpy.pi

            # create an image of 100 x 100 pixels and calculate the coordinates corresponding to every pixel
            centralpixel = self.config.centralpixel  # (column, row) = (delta, af)
            af_range = -app[1] * (numpy.arange(100) - centralpixel[1]) + af
            delta_range = app[0] * (numpy.arange(100) - centralpixel[0]) + delta

            # calculating the coordinates for simulating the image. This is only included
            # in this example for simulating of the images. It has no other use.

            k0 = 2 * numpy.pi / self.config.wavelength
            delta, af = numpy.meshgrid(delta_range, af_range)
            ai *= numpy.pi / 180
            delta *= numpy.pi / 180
            af *= numpy.pi / 180
            omega *= numpy.pi / 180

            qy = k0 * (
                numpy.cos(af) * numpy.cos(delta) - numpy.cos(ai) * numpy.cos(omega)
            )
            qx = k0 * (
                numpy.cos(af) * numpy.sin(delta) - numpy.cos(ai) * numpy.sin(omega)
            )
            qz = k0 * (numpy.sin(af) + numpy.sin(ai))

            # simulating the image
            data = (
                numpy.abs(
                    numpy.sin(qx * 10)
                    / numpy.sin(qx)
                    * numpy.sin(qy * 10)
                    / numpy.sin(qy)
                    * numpy.sin(qz * 10)
                    / numpy.sin(qz)
                )
                ** 2
            )
            weights = numpy.ones_like(data)

            yield data, weights, (self.config.wavelength, af, delta, omega, ai)

    def parse_config(self, config):
        """
        To collect and process data you need the values provided in the configuration file.
        These you can access locally through the provided config object. This is a dict with
        as keys the labels given in the configfile. To use them outside the parse_config method you attribute them
        to the self.config object which can be used throughout the input class. A warning will be
        generated afterwards for config values not popped out of the dict.
        """
        super(Input, self).parse_config(config)
        self.config.sdd = float(config.pop("sdd"))
        self.config.pixelsize = util.parse_tuple(
            config.pop("pixelsize"), length=2, type=float
        )
        self.config.centralpixel = util.parse_tuple(
            config.pop("centralpixel"), length=2, type=int
        )
        self.config.wavelength = float(config.pop("wavelength"))

    def get_destination_options(self, command):
        """
        Creates the arguments that you can use to construct an output filename. This method returns
        a dict object with keys that will can be  used in the configfile. In the configfile the output
        filename can now be described as 'destination = demo_{first}-{last}.hdf5'.
        This helps to organise the output automatically.
        """
        if not command:
            return False
        command = ",".join(command).replace(" ", ",")
        scans = util.parse_multi_range(command)
        return dict(first=min(scans), last=max(scans), range=",".join(command))



from typing import List, Optional, Union

import numpy
import matplotlib.colors
import matplotlib.cm
import mpl_toolkits.mplot3d

from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.lines import Line2D

from binoculars.space import Space

# Adapted from http://www.ster.kuleuven.be/~pieterd/python/html/plotting/interactive_colorbar.html
# which in turn is based on an example from http://matplotlib.org/users/event_handling.html


class DraggableColorbar(object):
    def __init__(self, cbar, mappable):
        self.cbar = cbar
        self.mappable = mappable
        self.press = None
        self.cycle = sorted(
            [i for i in dir(matplotlib.cm) if hasattr(getattr(matplotlib.cm, i), "N")]
        )

        try:  # matploltib 2.x
            cmap_name = cbar.get_cmap().name
        except AttributeError:  # matplotlib 3.x
            cmap_name = mappable.get_cmap().name

        self.index = self.cycle.index(cmap_name)
        self.canvas = self.cbar.patch.figure.canvas

    def connect(self):
        self.cidpress = self.canvas.mpl_connect("button_press_event", self.on_press)
        self.cidrelease = self.canvas.mpl_connect(
            "button_release_event", self.on_release
        )
        self.cidmotion = self.canvas.mpl_connect("motion_notify_event", self.on_motion)
        self.cidkeypress = self.canvas.mpl_connect("key_press_event", self.key_press)

    def disconnect(self):
        self.canvas.mpl_disconnect(self.cidpress)
        self.canvas.mpl_disconnect(self.cidrelease)
        self.canvas.mpl_disconnect(self.cidmotion)
        self.canvas.mpl_disconnect(self.cidkeypress)

    def on_press(self, event):
        if event.inaxes == self.cbar.ax:
            self.press = event.x, event.y

    def key_press(self, event):
        if event.key == "down":
            self.index += 1
        elif event.key == "up":
            self.index -= 1
        if self.index < 0:
            self.index = len(self.cycle)
        elif self.index >= len(self.cycle):
            self.index = 0
        cmap = self.cycle[self.index]
        self.mappable.set_cmap(cmap)
        self.cbar.patch.figure.canvas.draw()

    def on_motion(self, event):
        if self.press is None or event.inaxes != self.cbar.ax:
            return
        xprev, yprev = self.press
        # dx = event.x - xprev  # unused for now
        dy = event.y - yprev
        self.press = event.x, event.y

        if isinstance(self.cbar.norm, matplotlib.colors.LogNorm):
            scale = 0.999 * numpy.log10(self.cbar.norm.vmax / self.cbar.norm.vmin)
            if event.button == 1:
                self.cbar.norm.vmin *= scale ** numpy.sign(dy)
                self.cbar.norm.vmax *= scale ** numpy.sign(dy)
            elif event.button == 3:
                self.cbar.norm.vmin *= scale ** numpy.sign(dy)
                self.cbar.norm.vmax /= scale ** numpy.sign(dy)
        else:
            scale = 0.03 * (self.cbar.norm.vmax - self.cbar.norm.vmin)
            if event.button == 1:
                self.cbar.norm.vmin -= scale * numpy.sign(dy)
                self.cbar.norm.vmax -= scale * numpy.sign(dy)
            elif event.button == 3:
                self.cbar.norm.vmin -= scale * numpy.sign(dy)
                self.cbar.norm.vmax += scale * numpy.sign(dy)

        self.mappable.set_norm(self.cbar.norm)
        self.canvas.draw()

    def on_release(self, event):
        # force redraw on mouse release
        self.press = None
        self.mappable.set_norm(self.cbar.norm)
        self.canvas.draw()


def get_clipped_norm(data, clipping=0.0, log=True):
    if hasattr(data, "compressed"):
        data = data.compressed()
    else:
        data = data.flatten()

    if log:
        data = data[data > 0]
        if numpy.alen(data) == 0:
            return matplotlib.colors.LogNorm(1, 10)

    if clipping:
        chop = int(round(data.size * clipping))
        clip = sorted(data)[chop : -(1 + chop)]
        vmin, vmax = clip[0], clip[-1]
    else:
        vmin, vmax = data.min(), data.max()

    if log:
        return matplotlib.colors.LogNorm(vmin, vmax)
    else:
        return matplotlib.colors.Normalize(vmin, vmax)


def plot(
    space : Space,
    fig: Figure,
    ax : Axes,
    log: bool=True,
    loglog: bool=False,
    clipping: float=0.0,
    fit: Optional[bool]=None,
    norm: Optional[float]=None,
    colorbar: bool=True,
    labels: bool=True,
    interpolation: str="nearest",
    **plotopts
) -> Union[List[Line2D]]:
    if space.dimension == 1:
        data = space.get_masked()
        xrange = numpy.ma.array(space.axes[0][:], mask=data.mask)
        if fit is not None:
            if log:
                p1 = ax.semilogy(xrange, data, "wo", **plotopts)
                p2 = ax.semilogy(xrange, fit, "r", linewidth=2, **plotopts)
            elif loglog:
                p1 = ax.loglog(xrange, data, "wo", **plotopts)
                p2 = ax.loglog(xrange, fit, "r", linewidth=2, **plotopts)
            else:
                p1 = ax.plot(xrange, data, "wo", **plotopts)
                p2 = ax.plot(xrange, fit, "r", linewidth=2, **plotopts)
        else:
            if log:
                p1 = ax.semilogy(xrange, data, **plotopts)
            elif loglog:
                p1 = ax.loglog(xrange, data, **plotopts)
            else:
                p1 = ax.plot(xrange, data, **plotopts)
            p2 = []

        if labels:
            ax.set_xlabel(space.axes[0].label)
            ax.set_ylabel("Intensity (a.u.)")
        return p1 + p2

    elif space.dimension == 2:
        data = space.get_masked()

        # 2D IMSHOW PLOT
        xmin = space.axes[0].min
        xmax = space.axes[0].max
        ymin = space.axes[1].min
        ymax = space.axes[1].max

        if not norm:
            norm = get_clipped_norm(data, clipping, log)

        if fit is not None:
            im = ax.imshow(
                fit.transpose(),
                origin="lower",
                extent=(xmin, xmax, ymin, ymax),
                aspect="auto",
                norm=norm,
                interpolation=interpolation,
                **plotopts
            )
        else:
            im = ax.imshow(
                data.transpose(),
                origin="lower",
                extent=(xmin, xmax, ymin, ymax),
                aspect="auto",
                norm=norm,
                interpolation=interpolation,
                **plotopts
            )

        if labels:
            ax.set_xlabel(space.axes[0].label)
            ax.set_ylabel(space.axes[1].label)
        if colorbar:
            cbarwidget = fig.colorbar(im)
            fig._draggablecbar = DraggableColorbar(
                cbarwidget, im
            )  # we need to store this instance somewhere
            fig._draggablecbar.connect()
        return im

    elif space.dimension == 3:
        if not isinstance(ax, mpl_toolkits.mplot3d.Axes3D):
            raise ValueError(
                "For 3D plots, the 'ax' parameter must be an Axes3D instance (use for example gca(projection='3d') to get one)"
            )

        cmap = getattr(matplotlib.cm, plotopts.pop("cmap", "jet"))
        if norm is None:
            norm = get_clipped_norm(space.get_masked(), clipping, log)

        data = space.get()
        mask = numpy.bitwise_or(~numpy.isfinite(data), data == 0)
        gridx, gridy, gridz = tuple(grid[~mask] for grid in space.get_grid())
        im = ax.scatter(
            gridx,
            gridy,
            gridz,
            c=cmap(norm(data[~mask])),
            marker=",",
            alpha=0.7,
            linewidths=0,
        )

        # p1 = ax.plot_surface(gridx[0,:,:], gridy[0,:,:], gridz[0,:,:],  facecolors=cmap(norm(space.project(0).get_masked())), shade=False, cstride=1, rstride=1)
        # p2 = ax.plot_surface(gridx[:,-1,:], gridy[:,-1,:], gridz[:,-1,:], facecolors=cmap(norm(space.project(1).get_masked())), shade=False, cstride=1, rstride=1)
        # p3 = ax.plot_surface(gridx[:,:,0], gridy[:,:,0], gridz[:,:,0],  facecolors=cmap(norm(space.project(2).get_masked())), shade=False, cstride=1, rstride=1)

        if labels:
            ax.set_xlabel(space.axes[0].label)
            ax.set_ylabel(space.axes[1].label)
            ax.set_zlabel(space.axes[2].label)

        if fig._draggablecbar:
            fig._draggablecbar.disconnect()

        return im

    elif space.dimension > 3:
        raise ValueError(
            "Cannot plot 4 or higher dimensional spaces, use projections or slices to decrease dimensionality."
        )



# TODO: present exceptions based on errors.ExceptionBase in a gentle way to the user


class ExceptionBase(Exception):
    pass


class ConfigError(ExceptionBase):
    pass


class FileError(ExceptionBase):
    pass


class HDF5FileError(FileError):
    pass


class SubprocessError(ExceptionBase):
    pass


class BackendError(ExceptionBase):
    pass


class CommunicationError(ExceptionBase):
    pass


def addmessage(args, errormsg):
    if not args:
        arg0 = ""
    else:
        arg0 = args[0]
    arg0 += errormsg
    return (arg0,)
